/* Iria - An easy-customizing game, in the way of "Thunder Force IV".
 * Created by Juan Jos Ramrez. Synthetyx Software.
 *
 * gscn.c - Graphical Scenery Editor. An easy way for creating-editing
 * 	sceneries.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "gscn.h"

GraphicsContext physical, virtual;
void *expandedfont;
scn_header scenery;
imagepoint_list images;
eventpoint_list events;
unsigned long int point = 0, vpos = 0;

/* This function unloads all elements from memory */
int draw_object(int x, int y, event_record event) {
	FILE *f;
	obj_header header;
	png_bytep *image, aux, compiled;
	png_uint_32 width, height, i, j;
	imagepoint_list find;


	if (imagepoint_find(images, event.string)) {
		find = images;
		while (strcmp(find->filename, event.string))
			find = find->next;
		compiled = find->image;
		width = find->width;
		height = find->height;
	}
	else {
		// Reading object image...	
		if (!(f = fopen(event.string, "rb")))
			return FILE_DOESNT_EXIST;
		fread(&header, sizeof(obj_header), 1, f);
		fclose(f);

		// Reading image...
		if (readpng(header.image, &image, &width, &height) != GOODBYE_OK)
			return FILE_DOESNT_EXIST;
		if (!(aux = (png_bytep) malloc((width + 1) * (height + 1)))) {
			for (i = 0; i < height; i++)
				free(image[i]);
			free(image);
			return FILE_DOESNT_EXIST;
		}
		for (i = 0; i < width; i++)
			for (j = 0; j < height; j++)
				aux[i + (j * width)] = image[j][i];
		for (i = 0; i < height; i++)
			free(image[i]);
		free(image);
		if (!(compiled = (png_bytep) malloc((width + 3) * (height + 1)))) {
			free(aux);
			return FILE_DOESNT_EXIST;
		}
		gl_compileboxmask(width, height, aux, compiled);
		free(aux);
		if (imagepoint_insert(&images, point, event.string, compiled, width, height) != GOODBYE_OK) {
			free(compiled);
			return FILE_DOESNT_EXIST;
		}
	}
	gl_putboxmaskcompiled(x, y, width, height, compiled);
	return GOODBYE_OK;
}

/*
 * This function checks if the images are out of the screen. If so, we will
 * unload them.
 */
void check_images() {
	imagepoint_list aux, pre;

	pre = NULL;
	aux = images;
	while (aux)  
		if (((long int)aux->point < LEFT) || (aux->point > RIGHT)) {
			if (pre) {
				pre->next = aux->next;
				free(aux);
				aux = pre->next;
			}
			else {
				images = aux->next;
				free(aux);
				aux = images;
			}
		}
		else {
			pre = aux;
			aux = aux->next;
		}
}

/* This function carries out the object drawing */
void draw_event(int x, event_record event) {
	switch (event.type) {
		case UNCONDITIONAL_JUMP:	gl_printf(x, 0, "Unconditional jump to %lu", event.aux);
						gl_setpixel(x, 0, WHITE);
			break;
		case JUMP_WITH_COUNT:	gl_printf(x, 0, "Jump to %lu, count %lu", event.aux, event.count);
					gl_setpixel(x, 0, WHITE);	
			break;
		case JUMP_ENEMY_COUNT:	gl_printf(x, 0, "Jump to %lu with enemy count %lu", event.aux, event.count);
					gl_setpixel(x, 0, WHITE);		
			break;
		case SHOW_MESSAGE:	gl_printf(x, 0, "\"%s\"", event.string);
					gl_setpixel(x, 0, WHITE);		
			break;
		case CHANGE_PARAMETER:	gl_write(x, 0, "Change parameter"); 
					gl_setpixel(x, 0, WHITE);		
			break;
		case CREATE_OBJECT: if (draw_object(x, event.aux - vpos, event) != GOODBYE_OK)
								gl_printf(x, event.aux - vpos, "\"%s\"", event.string);
					gl_setpixel(x, event.aux - vpos, WHITE);		
			break;
	}
}

/* This function loads all events to memory */
int load_events(FILE *f) {
	event_record event;

	rewind(f);
	SKIP_HEADER
	while (fread(&event, sizeof(event_record), 1, f)) 
		if (eventpoint_insert(&events, event) != GOODBYE_OK)
			return NOT_ENOUGH_MEMORY;
	return GOODBYE_OK;
}		

/*
 * This function moves (if necesary) the screen according with the mouse 
 * position
 */
void adjust_screen() {
	// Horizontal adjust
	if ((mouse_getx() < THRESHOLD) && point)
		point -= SCREEN_SPEED;
	else if (WIDTH - mouse_getx() < THRESHOLD)
		point += SCREEN_SPEED;

	// Vertical adjust
	if ((mouse_gety() < THRESHOLD) && vpos)
		vpos -= SCREEN_SPEED;
	else if ((HEIGHT - mouse_gety() < THRESHOLD) && (vpos < scenery.height - HEIGHT))
		vpos += SCREEN_SPEED;
}

/* This function inserts, edits or deletes an event in the file */
void edit_insert_delete(FILE **file, char *filename, char action, event_record event, eventpoint_list old) {
	event_record aux;
	unsigned long int insert_position, pos;
	FILE *f;
	char end = 0;

	f = *file;
	switch (action) {
		case EDIT_EVENT:
				rewind(f);
				SKIP_HEADER
				if (!feof(f) && old) {
					do {
						fread(&aux, sizeof(event_record), 1, f);
					} while (memcmp(&aux, &old->node, sizeof(event_record)));
					fseek(f, -sizeof(event_record), SEEK_CUR);
					fwrite(&event, sizeof(event_record), 1, f);
					eventpoint_replace(events, aux, event);										
				}
			break;						
				
		case INSERT_EVENT:
				insert_position = find_file_position(f, event.point);
				fseek(f, 0, SEEK_END);	
				for (fgetpos(f, &pos); pos != insert_position; pos-=sizeof(event_record)) {
					fseek(f, pos - sizeof(event_record), SEEK_SET);
					fread(&aux, sizeof(event_record), 1, f);
					fseek(f, pos, SEEK_SET);
					fwrite(&aux, sizeof(event_record), 1, f);
				}
				fseek(f, insert_position, SEEK_SET);
				fwrite(&event, sizeof(event_record), 1, f);
				eventpoint_insert(&events, event);
			break;
		case DELETE_EVENT:
				rewind(f);
				SKIP_HEADER
				if (!feof(f) && old) {
						do {
							fread(&aux, sizeof(event_record), 1, f);
						} while (memcmp(&aux, &old->node, sizeof(event_record)));
						fseek(f, -sizeof(event_record), SEEK_CUR);
						while (!end) {
							fseek(f, sizeof(event_record), SEEK_CUR);
							end = fread(&aux, sizeof(event_record), 1, f) == 0;
							if (!end) {
								fseek(f, -2 * sizeof(event_record), SEEK_CUR);
								fwrite(&aux, sizeof(event_record), 1, f);
							}
						}
						fseek(f, -sizeof(event_record), SEEK_END);
						pos = ftell(f);
						fclose(*file);
						truncate(filename, pos);
						*file = fopen(filename, "rb+");
						eventpoint_delete(&events, old->node);
				}
	}
}

/* This function gets the event that you are clicking (if you are editing or deleting) if there are one */
eventpoint_list get_event(unsigned long int point, int y) {
	eventpoint_list aux, candidate = NULL;
	char finish = 0;
	
	aux = events;
	while (!finish && aux) {
		finish = !(aux->node.point < point);
		if (!finish)
			aux = aux->next;
	}
	finish = 0;
	while (aux && !finish)  {
		if (aux->node.point == point) {	// We found an event
			if ((aux->node.type == CREATE_OBJECT) && (aux->node.aux == y)) {	// That is the event we pointed
				candidate = aux;
				finish = TRUE;
			}
			else if (aux->node.type < CREATE_OBJECT)	// Possibly that is the event we pointed
				candidate = aux;
		}
		else finish = TRUE;
		aux = aux->next;
	}

	return candidate;
}
			

/* Main function: draws, edits, deletes, inserts... */
void graphevents(FILE **f, char *filename, event_record new_event, char action) {
	eventpoint_list aux;
	char button_released = 1;

	svga_initializations(&physical, &virtual, &expandedfont, 200);
	imagepoint_new(&images);	
	eventpoint_new(&events);
	mouse_setposition(WIDTH / 2, HEIGHT / 2);
	if (load_events(*f) != GOODBYE_OK) 
		exit(1);
	while (!keyboard_keypressed(SCANCODE_ESCAPE)) {
		check_images();
		aux = events;
		if (aux)
			while ((long int) aux->node.point < LEFT) {
				aux = aux->next;
				if (!aux)
					break;
			}
		if (aux)
			while (((long int) aux->node.point >= LEFT) && (aux->node.point <= RIGHT)) {
				draw_event(aux->node.point - LEFT, aux->node);
				aux = aux->next;
				if (!aux)
					break;
			}
		gl_line(WIDTH / 2, 0, WIDTH / 2, HEIGHT, 11);

		if (new_event.type == CREATE_OBJECT)
			new_event.aux = mouse_gety() + vpos;
		if (action == DELETE_EVENT) {
			gl_write(mouse_getx(), mouse_gety(), "Delete...");
			gl_setpixel(mouse_getx(), mouse_gety(), WHITE);			
		}
		else
			draw_event(mouse_getx(), new_event);
				
		if ((mouse_getbutton() == MOUSE_LEFTBUTTON) && button_released) {
			button_released = 0;
			new_event.point = point + mouse_getx() - WIDTH / 2;
			edit_insert_delete(f, filename, action, new_event, get_event(new_event.point, mouse_gety() + vpos));
		}
		else if (mouse_getbutton() != MOUSE_LEFTBUTTON)
			button_released = 1;
			

		if (new_event.type == CREATE_OBJECT)
			gl_printf(0, HEIGHT - 30, "Scenery point %lu\nX: %lu Y: %d", point, point + mouse_getx() - WIDTH / 2, new_event.aux);
		else
			gl_printf(0, HEIGHT - 30, "Scenery point %lu\nX: %lu", point, point + mouse_getx() - WIDTH / 2);

		
		vga_waitretrace();
		gl_copyscreen(&physical);
		gl_clearscreen(0);
		adjust_screen();
		mouse_update();
		keyboard_update();
	}
	imagepoint_destroy(&images);
	eventpoint_destroy(&events);
	svga_shutdown(&virtual, &expandedfont);	
}

/* This function inserts events in the scenery */
void insert_event(FILE *f, char *filename) {
	event_record data;
	
	fill_data(&data, NO_POINT);
	graphevents(&f, filename, data, INSERT_EVENT);
}

/* This function replaces events in the scenery */
void edit_event(FILE *f, char *filename) {
	event_record data;
	
	fill_data(&data, NO_POINT); // New event
	graphevents(&f, filename, data, EDIT_EVENT);
}

/* This function deletes events in the scenery */
void delete_event(FILE **f, char *filename) {
	event_record data;
	
	graphevents(f, filename, data, DELETE_EVENT);
}

/* This function loads the general data for an object */
void load_general(FILE *f) {
	int option;
	scn_header header;
	
	do {
		printf("What height do you want for the scenery?: ");
		scanf("%d", &header.height);
		printf("What speed do you want for the scenery?: ");
		scanf("%d", &header.speed);
		printf("What is the initial point for the scenery?: ");
		scanf("%lu", &header.init_point);
		do {
			printf("Choose a background effect for the scenery: \n"
				"1) None\n"
				"2) Stars\n"
				"3) Snow\n>");
			scanf("%d", &header.bg_type);
		} while ((header.bg_type < 1) || (header.bg_type > 3));
		printf("Type an image file for background (non-existing file for none): ");
		scanf("%s", header.filename);
		printf("Is all correct? [Y/N] ");
		do option = getchar(); while (option == 10);
	} while ((option != 'Y') && (option != 'y'));

	strcpy(header.version, SCENERY_VERSION);
	rewind(f);
	fwrite(&header, sizeof(scn_header), 1, f);
}

/* This function will load or create a scenery file */
void load_a_file(FILE **f, char *filename) {
	scn_header header;
	if (!(*f = fopen(filename, "rb+"))) {
		*f = fopen(filename, "wb+");
		printf("This is a new file, so I will ask you some questions.\nSee README.scenery for help.\n");		
		load_general(*f);
	}
	else {
		fread(&header, sizeof(scn_header), 1, *f);
		if (strcmp(header.version, SCENERY_VERSION)) {
			printf("This file has a different version (%s).\nNeeded version %s. Aborting...\n", header.version, SCENERY_VERSION);
			fclose(*f);
			exit(FILE_DOESNT_EXIST);
		}
	}
	rewind(*f);
	fread(&scenery, sizeof(scn_header), 1, *f);	// Read the global values
}

/* This function shows a menu and returns the choice */
int select_menu() {
	int option;

	do {
		printf("\nMAIN MENU\nSelect an option:\n"
			"1) Change global settings\n"
			"2) Edit an existing event\n"
			"3) Insert a new event\n"
			"4) Delete an event\n"
			"5) List events\n"
			"6) Go to a point (current %lu)\n"
			"7) Exit\n>", point);
		scanf("%d", &option);
	} while ((option < 1) || (option > SELECT_EXIT));
	return option;
}


int main(int argc, char **argv) {
	FILE *f;
	int option;

	if (argc < 2) {
		printf("You must specify a scenery file (.scn)\n");
		return FILE_DOESNT_EXIST;
	}
	
	printf("\tGRAPHICAL SCENERY EDITOR FOR IRIA. Synthetyx Software. See README.gscn\n");
	
	load_a_file(&f, argv[1]);

	
	while ((option = select_menu()) != SELECT_EXIT) 
		switch (option) {
			case CHANGE_GLOBALS: 
				load_general(f);
				break;
			case EDIT_EVENT:
				edit_event(f, argv[1]);
				break;
			case INSERT_EVENT:
				insert_event(f, argv[1]);
				break;
			case DELETE_EVENT:
				delete_event(&f, argv[1]);
				break;
			case LIST_EVENTS:
				list_events(f, "MAIN");
				break;
			case GOTO_POINT:
				printf("To what point? ");
				scanf("%lu", &point);
				break;
		}
	
	fclose(f);
	return GOODBYE_OK;
}
