/* Iria v0.3.5 - An easy-customizing game, in the way of "Thunder Force IV".
 * Created by Juan Jos Ramrez. Synthetyx Software.
 *
 * Iria.c - The main program for interpreting scenery and object files. 
 * 	This module can be understood as the game module.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include "iria.h"

scenery_struct scenery;	// Scenery definitions and events created by this scenery. See README.scenery and README.events for more details
object_struct  *object_index;	// Objects array with its events. See README.objects and README.events for more details
dead_list dead_objects;	// Objects that will be deleted
image_list images;	// All images used by objects
counter_list new_humans;	// On its creation, humans are inmortal and this counter is used for that.
counter_list lives;	// Lives of all humans created
int players = 2;	// Number of players
player_keys keys[2];
int frame_flag = 0;

static void frame(int sign) {
	frame_flag = 1;
}

void clock_initializations() {
	struct itimerval mytime;

	signal(SIGVTALRM, frame); // When the alarm signal is sent, we change the palette and set frame_flag
	mytime.it_interval.tv_sec = 0;
	mytime.it_interval.tv_usec = FPS;
	mytime.it_value.tv_sec = 0;
	mytime.it_value.tv_usec = FPS;
	setitimer(ITIMER_VIRTUAL, &mytime, NULL);
}

/* This function will add a new object to the objects list */
void create_object(object_list *objects, unsigned long int id, int posx, int posy, unsigned long int controled_by, object_list father) {
	object_node new;
	if	((controled_by <= players) && id) {
		new.type = id;
		if (!controled_by)
			new.x = posx;
		else {
			new.x = 0;
			counter_insert(&new_humans, controled_by, INMORTALITY_DURATION);
			counter_insert(&lives, controled_by, DEFAULT_LIVES);
		}			
		new.y = posy;
		new.image = object_index[id].image;
		new.controled_by = controled_by;
		weapon_new(&new.weapons);
		new.point = 0;
		new.visible = TRUE;
		new.life = object_index[id].life;
		new.energy = 0;
		new.father = father;
		new.next_event = object_index[id].events;
		counter_new(&new.counters);
		object_insert(objects, new, object_index);	
	}
}

/* This function will create an object (missile) if there is enough energy */
void shoot_object(object_list *objects, object_list shooter) {
	if (shooter->weapons)
		if (shooter->energy >= object_index[shooter->weapons->id].shoot_energy) { // There is enough energy
			create_object(objects, shooter->weapons->id, shooter->x + shooter->weapons->x, shooter->y + shooter->weapons->y, 0, shooter);
			shooter->energy = 0;	// Reset energy
		}
}

/*
 * This function draws background effects (like stars and snow) if they are 
 * enabled 
 */
int check_effects(int bg_type, int vpos, int height) {
	switch (bg_type) {
		case 1: return GOODBYE_OK;	// Nothing
				break;
		case 2: return draw_stars(vpos, height); // Stars background
				break;
	}
	return GOODBYE_OK;
}

/* This function will reset the 'next_event' pointer to the correct position */
void reset_eventlist(event_list *next_event, event_list init, unsigned long int point) {
	*next_event = init;
	while ((*next_event)->point < point) 
		*next_event = (*next_event)->next;	// We must reset the 'next_event' pointer when we jump to other point
}

/*
 * This function is very important. When the scenery or an object has reached
 * a point with an event, this function makes the changes that the event says
 */
void manage_event(event_node event, object_list *objects, int posx, unsigned long int *point, event_list init, event_list *next_event, object_list obj, counter_list *counters) {
	int result;

	switch (event.type) {
		case UNCONDITIONAL_JUMP:	*point = event.aux;
									reset_eventlist(next_event, init, *point);
				break;
		case JUMP_WITH_COUNT:	if (!counter_find(*counters, event.aux)) {
									result = counter_insert(counters, event.aux, event.count);
									EXIT_IF_NO_MEMORY
								}
								if (counter_value(*counters, event.aux) > 1) {
									counter_decrease(*counters, event.aux);
									*point = event.aux;
									reset_eventlist(next_event, init, *point);
								}
								else *next_event = (*next_event)->next;
				break;
		case CREATE_OBJECT:	create_object(objects, event.object_id, posx, (int) event.aux, event.count, obj);	// Create a new object
							*next_event = (*next_event)->next;
				break;
		case OBJECT_ERASE:	if (event.aux)	// With explosion
								create_object(objects, object_index[obj->type].die_id, obj->x, obj->y, obj->controled_by, NULL);
							dead_insert(&dead_objects, obj);
							*next_event = (*next_event)->next;
				break;
		case ADD_WEAPON:	weapon_insert(&obj->weapons, event.object_id, (int) event.aux, (int) event.count);
							*next_event = (*next_event)->next;
				break;
		default:	*next_event = (*next_event)->next;	// Unknown event type
				break;
	}
}

/* This function gets (and returns) the 'Adan' (first father) of the object */
object_list get_father(object_list obj) {
	while (obj->father) 
		obj = obj->father;
	return obj;
}

/* This function updates the 'father' field of each son which has lost a father :( */
void notify_sons(object_list objects, object_list father) {
	while (objects) {
		if (objects->father == father)
			objects->father = get_father(objects->father);	// New father
		objects = objects->next;
	}
}

/*
 * This function checks if the 2 objects are colliding. If so, both decreases 
 * life
 */
void check_collision(object_list obj1, object_list obj2, object_list *objects) {
	int x1, x2, y1, y2, init1, init2, colx, coly;
	png_bytep ptr1, ptr2;	// Pointers to images
	char collide = 0, end;
	int right1, right2, down1, down2;
	unsigned long int father_control1, father_control2;

	right1 = obj1->x + obj1->image->width;
	right2 = obj2->x + obj2->image->width;
	down1 = obj1->y + obj1->image->height;
	down2 = obj2->y + obj2->image->height;
	father_control1 = get_father(obj1)->controled_by;
	father_control2 = get_father(obj2)->controled_by;
	
	if ((!father_control1 && father_control2) || (father_control1 && !father_control2))	// They are enemies
	if ((object_index[obj1->type].damage && object_index[obj2->type].damage)) 
	if ((down1 > obj2->y) && (down2 > obj1->y))
	if ((right1 > obj2->x) && (right2 > obj1->x))
	if ((!object_index[obj1->type].inmortal || !object_index[obj2->type].inmortal)) 
	if (!counter_find(new_humans, obj1->controled_by) && !counter_find(new_humans, obj2->controled_by)) { // Any of these objects are 'flickering'
		ptr1 = obj1->image->image;
		ptr2 = obj2->image->image;
		init1 = obj2->y - obj1->y;
		init2 = -init1;
		// 'init' is the first horizontal line in obj2 at the same height that obj1

		for (y1 = 0; y1 < init1; y1++) 
			while (*ptr1++ != 0xff)  	// End of line
				ptr1 += *ptr1 + 1;
		
		for (y2 = 0; y2 < init2; y2++) 
			while (*ptr2++ != 0xff)  	// End of line
				ptr2 += *ptr2 + 1;
		
		for (; (y1 < obj1->image->height) && (y2 < obj2->image->height) && !collide; y1++, y2++) {
			x1 = obj1->x;
			x2 = obj2->x;
			while ((*ptr1 != 0xff) && (*ptr2 != 0xff) && !collide) {
				x2 += *ptr2++;	// Jump zeroes
				do {
					x1 += *ptr1++;	// Jump zeroes
					end = x1 > x2 + *ptr2;
					if (!end) 
						collide = x1 + *ptr1 >= x2;	// Collide, crossed lines
					if (collide) {
						colx = x1 > x2? x1 : x2;
						coly = y2 + obj2->y;
					}
					x1 += *ptr1;
					ptr1 += *ptr1 + 1;	// Jump nonzeroes
				} while (!end && (*ptr1 != 0xff) && !collide);
				x2 += *ptr2;
				ptr2 += *ptr2 + 1;	// Jump nonzeroes
			}
			while (*ptr1++ != 0xff) 
				ptr1 += *ptr1 + 1;	// Jump zeroes
			while (*ptr2++ != 0xff) 
				ptr2 += *ptr2 + 1;	// Jump zeroes
		} // for
		if (collide) {
			if (!object_index[obj1->type].inmortal) 
				obj1->life -= object_index[obj2->type].damage;		
			if (!object_index[obj2->type].inmortal)
				obj2->life -= object_index[obj1->type].damage;
			create_object(objects, object_index[obj1->type].hit_id, colx, coly, 0, obj1);
			create_object(objects, object_index[obj2->type].hit_id, colx, coly, 0, obj2);
		}
	} // huge 'if'
}

	

/* This function checks if the screen must go up, down or stay */
void adjust_vpos(object_list objects, int *vpos) {
	char someone_on_top = 0, someone_on_bottom = 0;	// There are someone on...?

	while (objects) {
		if (objects->controled_by != 0) {
			someone_on_top |= objects->y - *vpos < SCREEN_LIMIT;
			if (objects->image)
#ifdef USE_SVGALIB
				someone_on_bottom |= *vpos + HEIGHT - objects->y - objects->image->height < SCREEN_LIMIT;
			else
				someone_on_bottom |= *vpos + HEIGHT - objects->y < SCREEN_LIMIT;
#endif
		}
		objects = objects->next;
	}
	if (someone_on_top && !someone_on_bottom && (*vpos > 0))	// Screen must go up
		*vpos -= 4;
#ifdef USE_SVGALIB
	else if (!someone_on_top && someone_on_bottom && (*vpos + HEIGHT < scenery.height))	// Screen must go down
#endif
		*vpos += 4;
}

/* This function manages players actions (move, shoot, death...) */
void human_control(object_list *objects, object_list *human, int *vpos) {
	unsigned long int player, counter;
	int width, height;

	if ((*human)->image) {
		width = (*human)->image->width;
		height = (*human)->image->height;
	}
	else {
		width = 0;
		height = 0;
	}

	player = (*human)->controled_by - 1;
	(*human)->energy++;


	/* SHOOTING */	
	if (keyboard_keypressed(keys[player].KEYSHOOT)) {	// Shooting!
		shoot_object(objects, *human);
		if ((*human)->weapons)
			(*human)->image = object_index[(*human)->type].shootimage;
	}

	/* CHANGING WEAPON */
	if (keyboard_keypressed(keys[player].KEYCHANGE))
		if (keys[player].KEYRELEASED) 	// Change weapon
			if ((*human)->weapons) {	// Has weapons
				(*human)->weapons = (*human)->weapons->next;
				keys[player].KEYRELEASED = FALSE;
			}
			else;
		else;
	else 
		keys[player].KEYRELEASED = TRUE;

	/* MOVEMENT */
#ifdef USE_SVGALIB
	if ((keyboard_keypressed(keys[player].KEYRIGHT)) && ((*human)->x + width < WIDTH)) 
#endif
		(*human)->x += object_index[(*human)->type].move_aux;
	if ((keyboard_keypressed(keys[player].KEYLEFT)) && ((*human)->x > 0))
		(*human)->x -= object_index[(*human)->type].move_aux;
	if ((keyboard_keypressed(keys[player].KEYUP)) && ((*human)->y > *vpos)) {
		(*human)->y -= object_index[(*human)->type].move_aux;
		if (!keyboard_keypressed(keys[player].KEYDOWN))
			(*human)->image = object_index[(*human)->type].upimage;
	}
#ifdef USE_SVGALIB
	if ((keyboard_keypressed(keys[player].KEYDOWN)) && ((*human)->y + height < *vpos + HEIGHT)) {
#endif
		(*human)->y += object_index[(*human)->type].move_aux;
		if (!keyboard_keypressed(keys[player].KEYUP))
			(*human)->image = object_index[(*human)->type].downimage;
	}
	if (!keyboard_keypressed(keys[player].KEYUP) && !keyboard_keypressed(keys[player].KEYDOWN) && !(keyboard_keypressed(keys[player].KEYSHOOT) && (*human)->weapons))
		(*human)->image = object_index[(*human)->type].image;

	adjust_vpos(*objects, vpos);

	/* NEW HUMANS */
	if (counter_find(new_humans, (*human)->controled_by)) {
		if ((counter = counter_value(new_humans, (*human)->controled_by)) % 5)
			(*human)->image = NULL;
		if (counter)
			counter_decrease(new_humans, (*human)->controled_by);
		else
			counter_delete(&new_humans, (*human)->controled_by);
	}
}

/*
 * This function checks all objects for:
 *		- Moving them
 *		- Shooting
 *		- Decreases its life or destroy them
 */
void check_objects(object_list *objects, int *vpos) {
	object_list aux1, aux2;
	dead_list aux4;

	aux1 = *objects;
	while (aux1) {
		/* COLLISION CONTROL */
		aux2 = aux1->next;
		while (aux2) {
			if (aux1->image && aux2->image)
				check_collision(aux1, aux2, objects);
			aux2 = aux2->next;
		}
		if (aux1->controled_by == 0) {	// Machine controlled
			/* OBJECT MOVEMENT */
			switch(object_index[aux1->type].move_type) {
				case 3:	aux1->x -= object_index[aux1->type].move_aux;	// To left
					break;
				case 4:	aux1->x += object_index[aux1->type].move_aux;	// To right
					break;
			}
			/* OBJECT SHOOTING */
			aux1->energy++;
			switch(object_index[aux1->type].shoot_type) {
				case 1: shoot_object(objects, aux1); // Always
					break;
			}
		}
		else 
			human_control(objects, &aux1, vpos);
		/* EVENTS CONTROL */
		if (aux1->next_event) while (aux1->next_event->point == aux1->point) { 
			manage_event(*aux1->next_event, objects, aux1->x, &aux1->point, object_index[aux1->type].events, &aux1->next_event, aux1, &aux1->counters);
			if (!aux1->next_event)
				break;
		}
		/* ADVANCE POSITION */
		aux1->point++;
		aux1 = aux1->next;
	}	
	/* OBJECTS ARE DEAD? */
	aux1 = *objects;
	while (aux1) {
#ifdef USE_SVGALIB
		if ((aux1->life < 1) && !object_index[aux1->type].inmortal)   // The object is dead
#endif
			dead_insert(&dead_objects, aux1);	// It will be deleted
		aux1 = aux1->next;
	}
	while (dead_objects) {	// We will delete all objects which were notified to be dead
		aux1 = dead_objects->object;
		create_object(objects, object_index[aux1->type].die_id, aux1->x, aux1->y, 0, NULL);	// Creating the final explosion
		if (!aux1->controled_by) {
			notify_sons(*objects, aux1);	// It sons must know that their father has been died ;)
			object_delete(objects, aux1);	// Delete finally
		}
		else if (counter_value(lives, aux1->controled_by)) {	// There are more lives
			counter_decrease(lives, aux1->controled_by);
			if(aux1->weapons)
				weapon_delete(&aux1->weapons);
			aux1->life = object_index[aux1->type].life;	// Reincarnation
			counter_insert(&new_humans, aux1->controled_by, INMORTALITY_DURATION); 
		}
		else {
			notify_sons(*objects, aux1);	// It sons must know that their father has been died ;)
			object_delete(objects, aux1);	// Delete finally
		}
		aux4 = dead_objects;
		dead_objects = dead_objects->next;	
		free (aux4);	// Erases this node
	}
}


/* This function manages the current scenery point and increases the counter */
int check_scenery(scenery_struct *s, object_list *objects) {
	if (s->next_event) while (s->next_event->point == s->current_point) {
#ifdef USE_SVGALIB
		manage_event(*s->next_event, objects, WIDTH, &s->current_point, s->events, &s->next_event, NULL, &s->counters);
#endif
		if (!s->next_event) // We've reached the end of the scenery
			return SCENERY_END;
	}
	s->current_point++;
	return GOODBYE_OK;
}

/* This function draws all objects in object list */
void draw_objects(object_list objects, int vpos) {
	while (objects) {
		if (objects->image && objects->visible)
#ifdef USE_SVGALIB
			gl_putboxmaskcompiled(objects->x, objects->y - vpos, objects->image->width, objects->image->height, (void *) objects->image->image);
#endif
		objects = objects->next;
	}
}

/*
 * This function is the main function. It makes graphics initializations, 
 * runs scenery...
 */
int play_game() {
	int result = GOODBYE_OK;
	object_list objects;	// Objects alive
#ifdef USE_SVGALIB
	GraphicsContext physicalscreen, virtualscreen;
	void *expandedfont;
#endif

#ifdef USE_SVGALIB
	svga_initializations(&physicalscreen, &virtualscreen, &expandedfont, scenery.height);
#endif
	clock_initializations();	// Set the clock counter for changing palette and set fps
	object_new(&objects);
	dead_new(&dead_objects);
	counter_new(&new_humans);
	counter_new(&lives);
#ifdef USE_SVGALIB
	while (!keyboard_keypressed(SCANCODE_ESCAPE) && (result == GOODBYE_OK)) {
#endif
		result = check_effects(scenery.bg_type, scenery.vpos, scenery.height);	
		EXIT_IF_NO_MEMORY
		check_objects(&objects, &scenery.vpos);
		result = check_scenery(&scenery, &objects);
		draw_objects(objects, scenery.vpos);

#ifdef USE_SVGALIB		
		vga_waitretrace();gl_copyscreen(&physicalscreen);gl_clearscreen(0);// Draw objects
		keyboard_update();
#endif
		while (!frame_flag)
			{}
		frame_flag = 0;
	}
	destroy_stars();
	object_destroy(&objects);
	counter_destroy(&new_humans);
	counter_destroy(&lives);
#ifdef USE_SVGALIB
	svga_shutdown(&virtualscreen, &expandedfont);
#endif

	return !GAME_OVER;
}
		
void show_usage() {
	printf("You doesn't specified any scenery file\n"
			"Please, use \"iria file1.scn [file2.scn ...]\"\n");
	exit(USAGE_ERROR);
}

int main(int argc, char **argv) {
	int i, result, game_result = !GAME_OVER;
	string_list object_files;
	imagefile_list images_loaded;
	unsigned long int number_of_objects;

	if (argc < 2)
		show_usage();
	printf("\n\n\n\tWelcome to Iria %s by Juan Jos Ramrez (Synthetyx Software)\n\t\t\t\t\t^_^\n\n", IRIA_VERSION);
	sleep(1);

	keys[0].KEYUP = SCANCODE_CURSORBLOCKUP;
	keys[0].KEYDOWN = SCANCODE_CURSORBLOCKDOWN;
	keys[0].KEYRIGHT = SCANCODE_CURSORBLOCKRIGHT;
	keys[0].KEYLEFT = SCANCODE_CURSORBLOCKLEFT;
	keys[0].KEYSHOOT = SCANCODE_SPACE;
	keys[0].KEYCHANGE = SCANCODE_ENTER;

	keys[1].KEYUP = SCANCODE_W;
	keys[1].KEYDOWN = SCANCODE_S;
	keys[1].KEYRIGHT = SCANCODE_D;
	keys[1].KEYLEFT = SCANCODE_A;
	keys[1].KEYSHOOT = SCANCODE_O;
	keys[1].KEYCHANGE = SCANCODE_P;

	for (i = 1; (i < argc) && (game_result != GAME_OVER); i++) {
		string_new(&object_files);
		imagefile_new(&images_loaded);
		image_new(&images);		
		result = which_files_I_need(argv[i], &object_files, &number_of_objects);
		if (result == FILE_DOESNT_EXIST) {
			printf("Some files doesn't exist or are unreadable, aborting...\n");
			exit(FILE_DOESNT_EXIST);
		}
		else EXIT_IF_NO_MEMORY
		result = load_scenery(argv[i], &scenery, object_files, &images, &images_loaded);
		EXIT_IF_NO_MEMORY
		result = load_objects(object_files, &object_index, &images, &images_loaded);
		EXIT_IF_NO_MEMORY

		unload_file_lists(&object_files, &images_loaded);

		game_result = play_game();

		unload_game_elements(&scenery, &object_index, &images, number_of_objects);
		
	}
		return GOODBYE_OK;
}
