/* Iria - An easy-customizing game, in the way of "Thunder Force IV".
 * Created by Juan Jos Ramrez. Synthetyx Software.
 *
 * lists.c - A module with a lot of lists definitions and functions.
 *
 */

#include "lists.h"

/*****************************************************************************
 *                     STRING LISTS SECTION                                  *
 *****************************************************************************/

/* Every list needs to be initialized. This function inits a new list */
void string_new(string_list *l) {
	*l = NULL;
}

/* This function adds an element to the list */
int string_insert(string_list *l, char *s) {
	string_list aux;
	aux = (string_list) malloc(sizeof(string_node));
	if (!aux)
		return NOT_ENOUGH_MEMORY;
	strcpy(aux->info, s);
	aux->next = *l;
	*l = aux;
	return GOODBYE_OK;
}

/* This function extracts an element from the list */
void string_extract(string_list *l, char *s) {
	string_list aux;
	if (*l) {
		aux = *l;
		strcpy(s, (*l)->info);
		*l = (*l)->next;
		free(aux);
	}
	else printf("Warning: You tried to extract from NULL list\n");
}

/* This function unloads all elements from memory */
void string_destroy(string_list *l) {
	string_list aux;
	while (*l) {
		aux = *l;
		*l = (*l)->next;
		free(aux);
	}
}

/*
 * This functions returns 1 if the string is equal to one or more elements of
 * the list. Else, it returns 0
 */
int string_find(string_list l, char *s) {
	int found = 0;
	while (l && !found) {
		found = !strcmp(l->info, s);
		l = l->next;
	}
	return found;
}

/*****************************************************************************
 *                     IMAGE FILE LIST SECTION                               *
 *****************************************************************************/

/* This function inits a new image file list */
void imagefile_new(imagefile_list *l) {
	*l = NULL;
}

/* This function inserts a new image file node */
int imagefile_insert(imagefile_list *l, char *filename, image_list image) {
	imagefile_list aux;

	aux = (imagefile_list) malloc(sizeof(imagefile_node));
	if (!aux)
		return NOT_ENOUGH_MEMORY;
	strcpy(aux->filename, filename);
	aux->image = image;
	aux->next = *l;
	*l = aux;
	return GOODBYE_OK;
}

/* This function unloads all elements from memory */
void imagefile_destroy(imagefile_list *l) {
	imagefile_list aux;
	while (*l) {
		aux = *l;
		*l = (*l)->next;
		free(aux);
	}
}

/*
 * This functions returns 1 if the filename is equal to one or more filenames of
 * the list. Else, it returns 0
 */
int imagefile_find(imagefile_list l, char *filename) {
	int found = 0;
	while (l && !found) {
		found = !strcmp(l->filename, filename);
		l = l->next;
	}
	return found;
}

/*****************************************************************************
 *                     IMAGE POINT LIST SECTION                              *
 *****************************************************************************/

/* This function inits a new image point list */
void imagepoint_new(imagepoint_list *l) {
	*l = NULL;
}

/* This function inserts a new image point node */
int imagepoint_insert(imagepoint_list *l, unsigned long int point, char *filename, png_bytep image, png_uint_32 width, png_uint_32 height) {
	imagepoint_list aux;

	aux = (imagepoint_list) malloc(sizeof(imagepoint_node));
	if (!aux)
		return NOT_ENOUGH_MEMORY;
	strcpy(aux->filename, filename);
	aux->point = point;
	aux->image = image;
	aux->width = width;
	aux->height = height;
	aux->next = *l;
	*l = aux;
	return GOODBYE_OK;
}

/* This function unloads all elements from memory */
void imagepoint_destroy(imagepoint_list *l) {
	imagepoint_list aux;
	while (*l) {
		aux = *l;
		*l = (*l)->next;
		free(aux);
	}
}

/*
 * This functions returns 1 if the filename is equal to one or more filenames of
 * the list. Else, it returns 0
 */
int imagepoint_find(imagepoint_list l, char *filename) {
	int found = 0;
	while (l && !found) {
		found = !strcmp(l->filename, filename);
		l = l->next;
	}
	return found;
}

/*****************************************************************************
 *                    EVENT POINT LISTS SECTION                              *
 *****************************************************************************/

/* This function inits a new list */
void eventpoint_new(eventpoint_list *l) {
	*l = NULL;
}

/* This function loads a new element to the list, ordered by point */
int eventpoint_insert(eventpoint_list *l, event_record node) {
	eventpoint_list pre, aux;
	
	pre = NULL;
	aux = *l;
	if (aux)
		while (node.point > aux->node.point) {
			pre = aux;
			aux = aux->next;
			if (!aux)
				break;
		}
	aux = (eventpoint_list) malloc(sizeof(eventpoint_node));
	if (!aux)
		return NOT_ENOUGH_MEMORY;
	aux->node = node;
	strcpy(aux->node.string, node.string);
	if (pre) {
		aux->next = pre->next;
		pre->next = aux;
	}
	else {
		aux->next = *l;
		*l = aux;
	}
	return GOODBYE_OK;
}

/* This function replaces an element */
void eventpoint_replace(eventpoint_list l, event_record old, event_record new) {
	
	while (memcmp(&l->node, &old, sizeof(event_record))) {	// I asume that the element exists
		l = l->next;
	}
	l->node = new;
}

/* This function deletes an element */
void eventpoint_delete(eventpoint_list *l, event_record event) {
	eventpoint_list aux, pre;
	
	pre = NULL;
	aux = *l;
	while (memcmp(&aux->node, &event, sizeof(event_record))) {	// I asume that the element exists
		pre = aux;
		aux = aux->next;
	}
	if (pre) 
		pre->next = aux->next;
	else
		*l = aux->next;
	free(aux);
}

/* This function deletes all elements from the list */
void eventpoint_destroy(eventpoint_list *l) {
	eventpoint_list aux;

	while (*l) {
		aux = *l;
		*l = aux->next;
		free(aux);
	}
}

/*****************************************************************************
 *                       EVENT LISTS SECTION                                 *
 *****************************************************************************/

/* This function initializes a new event list */
void event_new(event_list *l) {
	*l = NULL;
}

/* This function inserts a new element to the end of the list */
int event_insert(event_list *l, event_record data, void *pointer, unsigned long int object_id) {
	event_list aux, aux2;
	aux = (event_list) malloc(sizeof(event_node));
	if (!aux)
		return NOT_ENOUGH_MEMORY;
	aux->type = data.type;
	strcpy(aux->string, data.string);
	aux->pointer = pointer;
	aux->object_id = object_id;
	aux->point = data.point;
	aux->aux = data.aux;
	aux->count = data.count;
	aux->next = NULL;
	aux2 = *l;
	if (aux2) {
		while (aux2->next)
			aux2 = aux2->next;		
		aux2->next = aux;
	}
	else *l = aux;
	return GOODBYE_OK;
}

/* This function unloads all elements from memory */
void event_destroy(event_list *l) {
	event_list aux;
	while (*l) {
		aux = *l;
		*l = (*l)->next;
		free(aux);
	}
}

/*****************************************************************************
 *                      IMAGE LISTS SECTION                                  *
 *****************************************************************************/

/* This function inits a new image list */
void image_new(image_list *l) {
	*l = NULL;
}

/* This function inserts a new element to the image list */
int image_insert(image_list *l, png_bytep image, png_uint_32 width, png_uint_32 height) {
	image_list aux;
	aux = (image_list) malloc(sizeof(image_node));
	if (!aux)
		return NOT_ENOUGH_MEMORY;
	aux->image = image;
	aux->width = width;
	aux->height = height;
	aux->next = *l;
	*l = aux;
	return GOODBYE_OK;
}

/* This function unloads all elements from memory */
void image_destroy(image_list *l) {
	image_list aux;
	while (*l) {
		aux = *l;
		*l = (*l)->next;
		free(aux);
	}
}

/*****************************************************************************
 *                       COUNTER LISTS SECTION                               *
 *****************************************************************************/

/* This function will create a new list */
void counter_new(counter_list *l) {
	*l = NULL;
}

/* This will insert a new element */
int counter_insert(counter_list *l, unsigned long int point, unsigned long int count) {
	counter_list aux;
	aux = (counter_list) malloc(sizeof(counter_node));
	if (!aux)
		return NOT_ENOUGH_MEMORY;
	aux->point = point;
	aux->count = count;
	aux->next = *l;
	*l = aux;
	return GOODBYE_OK;
}
	
/*
 * This functions returns 1 if point is found in the list.
 * Else, it returns 0
 */
int counter_find(counter_list l, unsigned long int point) {
	int found = 0;
	while (l && !found) {
		found = l->point == point;
		l = l->next;
	}
	return found;
}

/* This function will returns the value of the counter in a point */
unsigned long int counter_value(counter_list l, unsigned long int point) {
	while (l->point != point)
		l = l->next;
	return l->count;	// I assume that the point exists
}

/* This function will decrease in 1 unit the value of the counter */
void counter_decrease(counter_list l, unsigned long int point) {
	while (l->point != point)
		l = l->next;
	l->count--;
}

/* This function looks for an element and deletes it */
void counter_delete(counter_list *l, unsigned long int point) {
	counter_list aux, pre;

	aux = *l;
	pre = NULL;
	while (aux->point != point) {				// I assume that the point exists
		pre = aux;
		aux = aux->next;
	}
	if (pre)
		pre->next = aux->next;
	else
		*l = aux->next;
	free(aux);
}

/* This function deletes all elements from the list */
void counter_destroy(counter_list *l) {
	counter_list aux;
	while (*l) {
		aux = *l;
		*l = (*l)->next;
		free(aux);
	}
}

/*****************************************************************************
 *                     WEAPON LISTS SECTION                                  *
 *****************************************************************************/

/* This function inits a new list */
void weapon_new(id_list *l) {
	*l = NULL;
}

/* This function inserts a new element to the list */
int weapon_insert(id_list *l, unsigned long int id, int x, int y) {
	id_list aux;

	aux = (id_list) malloc(sizeof(id_node));
	if (!aux)
		return NOT_ENOUGH_MEMORY;
	aux->id = id;
	aux->x = x;
	aux->y = y;
	if (*l) {
		aux->next = (*l)->next;
		(*l)->next = aux;
	}
	else {
		aux->next = aux;
		*l = aux;
	}
	return GOODBYE_OK;
}

/* This function deletes the first weapon */
void weapon_delete(id_list *l) {
	id_list aux, pre;

	aux = *l;
	if (*l == aux->next)
		*l = NULL;
	else {
		pre = aux;
		while (pre->next != *l)
			pre = pre->next;
		*l = aux->next;
		pre->next = *l;
	}
	free(aux);
}

/* This function deletes all elements from the list */
void weapon_destroy(id_list *l) {
	id_list init, aux;

	init = *l;
	if (*l)
		do {
			aux = *l;
			*l = aux->next;
			free(aux);
		} while (*l != init);
}

/*****************************************************************************
 *                     OBJECT LISTS SECTION                                  *
 *****************************************************************************/

/* This function inits a new list */
void object_new(object_list *l) {
	*l = NULL;
}

/* This function inserts a new element to the list, ordered by layer (from object_index) */
int object_insert(object_list *l, object_node data, object_struct *object_index) {
	object_list pre, aux;
	
	pre = NULL;
	aux = *l;
	if (aux)
		while (object_index[aux->type].layer < object_index[data.type].layer) {
			pre = aux;
			aux = aux->next;
			if (!aux)
				break;
		}
	aux = (object_list) malloc(sizeof(object_node));
	if (!aux)
		return NOT_ENOUGH_MEMORY;
	*aux = data;
	if (pre) {
		aux->next = pre->next;
		pre->next = aux;
	}
	else {
		aux->next = *l;
		*l = aux;
	}
	return GOODBYE_OK;
}
/*
int object_insert(object_list *l, object_node data) {
	object_list aux;
	aux = malloc(sizeof(object_node));
	*aux = data;
	aux->next = *l;
	*l = aux;
	return GOODBYE_OK;
}
*/
/* This function finds and deletes an element from the list */
void object_delete(object_list *l, object_list element) {
	object_list pre, aux;

	pre = NULL;
	aux = *l;
	while (aux != element) {	// I suposse that the element exists
		pre = aux;
		aux = aux->next;
	}
	if (pre) 
		pre->next = aux->next;
	else 	// First element
		*l = aux->next;
	counter_destroy(&aux->counters);
	weapon_destroy(&aux->weapons);
	free(aux);
}

/* This function deletes all elements from the list */
void object_destroy(object_list *l) {
	object_list aux;
	while (*l) {
		aux = *l;
		*l = (*l)->next;
		counter_destroy(&aux->counters);		
		weapon_destroy(&aux->weapons);
		free(aux);
	}
}


/*****************************************************************************
 *                     DEAD (OBJECTS) LISTS SECTION                          *
 *****************************************************************************/

/* This function inits a new list */
void dead_new(dead_list *l) {
	*l = NULL;
}

/* This function inserts a new element to the list */
int dead_insert(dead_list *l, object_list new) {
	dead_list aux;

	aux = (dead_list) malloc(sizeof(dead_node));
	if (!aux)
		return NOT_ENOUGH_MEMORY;
	aux->object = new;
	aux->next = *l;
	*l = aux;
	return GOODBYE_OK;
}

/* This function deletes all elements from the list */
void dead_destroy(dead_list *l) {
	dead_list aux;

	while (*l) {
		aux = *l;
		*l = aux->next;
		free(aux);
	}
}
