/* Iria - An easy-customizing game, in the way of "Thunder Force IV".
 * Created by Juan Jos Ramrez. Synthetyx Software.
 *
 * loadfiles.c - A module for loading scenery, objects and image files.
 *
 */

#include "iria.h"

/*
 * This function returns the position of the file in the array. That's the
 * object id.
 */
unsigned long int find_object_id(char *filename, string_list files) {
	char found = 0;
	unsigned long int count = 0;
	
	while (files && !found) {
		found = !strcmp(files->info, filename);
		count++;
		files = files->next;
	}
	if (!found)
		return 0;
	else
		return count;
}

/*
 * This function returns a pointer to the image bitmap or NULL if there are 
 * errors. If the image is a compiled mask image, we will compile it
 */
image_list find_image_id(char *filename, image_list *images, char compiled, imagefile_list *images_loaded) {
	png_bytep *image, final1, final2;
	png_uint_32 width, height, i, j;
	imagefile_list aux;

	if (imagefile_find(*images_loaded, filename)) {
		aux = *images_loaded;
		while (strcmp(aux->filename, filename)) 
			aux = aux->next;
		return aux->image;
	}
	if (readpng(filename, &image, &width, &height) != GOODBYE_OK)
		return NULL;
	if (!(final1 = (png_bytep) malloc((width + 1) * (height + 1)))) {
		for (i = 0; i < height; i++)
			free(image[i]);
		free(image);
		return NULL;
	}
	for (i = 0; i < width; i++)
		for (j = 0; j < height; j++)
			final1[i + (j * width)] = image[j][i];	
	for (i = 0; i < height; i++)
		free(image[i]);
	free(image);
	if (compiled) {
		if (!(final2 = (png_bytep) malloc((width + 3) * (height + 1)))) {
			free(final1);
			return NULL;
		}
#ifdef USE_SVGALIB
		gl_compileboxmask(width, height, final1, final2);
#endif
		free(final1);
		image_insert(images, final2, width, height);
		imagefile_insert(images_loaded, filename, *images);
		return *images;
	}
	else {
		image_insert(images, final2, width, height);
		imagefile_insert(images_loaded, filename, *images);
		return *images;
	}
}

/*
 * This function finds out which files I need. Since every .scn or .iri files
 * has several references to other files, I need to explore them
 */
int which_files_I_need(char *scn_filename, string_list *files, unsigned long int *number_of_objects) {
	FILE *f, *check;
	event_record data;	
	obj_header fileheader;
	char current[STRING_LENGTH];
	string_list loaded; // Filenames loaded from a file
	
	string_new(&loaded);
	if (!(f = fopen(scn_filename, "rb"))) {
		printf("File \"%s\" doesnt exist or it is unreadable!\n", scn_filename);
		return FILE_DOESNT_EXIST;
	}
	SKIP_SCN_HEADER
	while (fread(&data, sizeof(event_record), 1, f)) { // We read all the scn file looking for new objects
		if (data.type == 6) // Insert an object
				if ((check = fopen(data.string, "rb"))) {
					if (!string_find(loaded, data.string)) // We don't want repeated elements
						if (string_insert(&loaded, data.string) == NOT_ENOUGH_MEMORY)
							return NOT_ENOUGH_MEMORY;
					fclose(check);
				}
	}
	fclose(f);
	while (loaded) {
		string_extract(&loaded, current);
		if ((f = fopen(current, "rb"))) {
			if (!string_find(*files, current)) {	// We don't want repeated elements 
				if (string_insert(files, current) == NOT_ENOUGH_MEMORY)
					return NOT_ENOUGH_MEMORY;
				fread(&fileheader, sizeof(obj_header), 1, f);
				if (fileheader.damage) {
					if (!string_find(loaded, fileheader.hit_file)) // We don't want repeated elements
						if (string_insert(&loaded, fileheader.hit_file) == NOT_ENOUGH_MEMORY) // shoot file is a new object to load
							return NOT_ENOUGH_MEMORY;
					if (!string_find(loaded, fileheader.die_file)) // We don't want repeated elements
						if (string_insert(&loaded, fileheader.die_file) == NOT_ENOUGH_MEMORY) // shoot file is a new object to load
							return NOT_ENOUGH_MEMORY;
				}
				if (!string_find(loaded, fileheader.goodies)) // We don't want repeated elements
					if (string_insert(&loaded, fileheader.goodies) == NOT_ENOUGH_MEMORY) // Goodies object
						return NOT_ENOUGH_MEMORY;
				SKIP_OBJ_HEADER
				while (fread(&data, sizeof(event_record), 1, f))
					if ((data.type == CREATE_OBJECT) || ((data.type == CHANGE_SHOOT) && (data.aux != 6)) || (data.type == ADD_WEAPON))  // Create object OR shoot file != none OR Add a weapon object
							if (!string_find(loaded, data.string))	// We don't want repeated elements
								if (string_insert(&loaded, data.string) == NOT_ENOUGH_MEMORY)
									return NOT_ENOUGH_MEMORY;
			}
			fclose(f);
		}
		else 
			printf("Warning: File \"%s\" doesnt exist or it isn't readable!\n", current);
	}
	loaded = *files;
	for (*number_of_objects = 0; loaded; (*number_of_objects)++)
		loaded = loaded->next;
	return GOODBYE_OK;
}

/* This function inserts events into the list */
int load_events(FILE *f, event_list *l, string_list files, image_list *images, imagefile_list *images_loaded) {
	event_record data;
	void *pointer = NULL;
	unsigned long int object_id;

	while (fread(&data, sizeof(event_record), 1, f)) {
		if ((data.type == CHANGE_PARAMETER) && (data.aux == 1)) // Change background image
			pointer = find_image_id(data.string, images, 0, images_loaded);
		else if (data.type == CREATE_OBJECT) // Insert a new object
			object_id = find_object_id(data.string, files);
		else if (data.type == CHANGE_OBJECT_IMAGE) // Change object image 
			pointer = find_image_id(data.string, images, 1, images_loaded);
		else if ((data.type == CHANGE_MOVEMENT) && (data.aux == 1)) // Change movement behaviour with file specification
			pointer = NULL; // In future, it will be a pointer to a movement list
		else if ((data.type == CHANGE_SHOOT) && (data.aux != 5)) // Change shoot behaviour, a new object will be inserted
			object_id = find_object_id(data.string, files);
		else if (data.type == ADD_WEAPON)
			object_id =find_object_id(data.string, files);

		if (event_insert(l, data, pointer, object_id) == NOT_ENOUGH_MEMORY)
			return NOT_ENOUGH_MEMORY;
	}
	return GOODBYE_OK;
}

/* This function loads the scenery file to memory */
int load_scenery(char *filename, scenery_struct *s, string_list files, image_list *images, imagefile_list *images_loaded) {
	FILE *f;
	scn_header header;

	printf("Loading %s...\n", filename);
	if (!(f = fopen(filename, "rb")))
		return FILE_DOESNT_EXIST;
	rewind(f);	
	fread(&header, sizeof(scn_header), 1, f);
	s->height = header.height;
	s->vpos = 0;
	s->bg_type = header.bg_type;
	s->speed = header.speed;
	s->current_point = header.init_point;
	counter_new(&s->counters);
	s->back_image = find_image_id(header.filename, images, 0, images_loaded);
	event_new(&s->events);
	if (load_events(f, &s->events, files, images, images_loaded) == NOT_ENOUGH_MEMORY)
		return NOT_ENOUGH_MEMORY;
	s->next_event = s->events;
	fclose(f);
	return GOODBYE_OK;
}

/*
 * This function loads all objects into an array. The array index is the
 * object id. 
 */
int load_objects(string_list files, object_struct **objects, image_list *images, imagefile_list *images_loaded) {
	FILE *f;
	unsigned long int count = 0;
	string_list aux;
	obj_header header;
	
	aux = files;
	while (aux) {
		count++; // I need to know how many objects will be loaded from an scenery file
		aux = aux->next;
	}
	if (!(*objects = (object_struct *) malloc((count + 1) * sizeof(object_struct))))
		return NOT_ENOUGH_MEMORY;
	for (count = 1, aux = files; aux; count++, aux = aux->next) {
		printf("Loading %s...\n", aux->info);
		f = fopen(aux->info, "rb");
		fread(&header, sizeof(obj_header), 1, f);
		(*objects)[count].image = find_image_id(header.image, images, 1, images_loaded);
		(*objects)[count].shootimage = find_image_id(header.shootimage_file, images, 1, images_loaded);
		(*objects)[count].upimage = find_image_id(header.upimage_file, images, 1, images_loaded);
		(*objects)[count].downimage = find_image_id(header.downimage_file, images, 1, images_loaded);
		(*objects)[count].inmortal = header.inmortal;
		(*objects)[count].life = header.life;
		(*objects)[count].layer = header.layer;
		(*objects)[count].damage = header.damage;
		(*objects)[count].move_type = header.move_type;
		strcpy((*objects)[count].move_file, header.move_file);
		(*objects)[count].move_aux = header.move_aux;
		(*objects)[count].shoot_type = header.shoot_type;
		(*objects)[count].shoot_count = header.shoot_count;
		(*objects)[count].shoot_energy = header.shoot_energy;
		(*objects)[count].hit_id = find_object_id(header.hit_file, files);
		(*objects)[count].die_id = find_object_id(header.die_file, files);
		(*objects)[count].goodies = find_object_id(header.goodies, files);
		event_new(&(*objects)[count].events);
		load_events(f, &(*objects)[count].events, files, images, images_loaded);
		fclose(f);
	}
	return GOODBYE_OK;
}
