/* Iria - An easy-customizing game, in the way of "Thunder Force IV".
 * Created by Juan Jos Ramrez. Synthetyx Software.
 *
 * structs.h - Some file headers and event records definitions
 *
 */

#ifndef __STRUCTS_DEFINED
#define __STRUCTS_DEFINED

#include <png.h>
#define STRING_LENGTH 128

/* Scenery files header */
typedef struct {
	char version[10];	// Version of the scenery file
	int height, bg_type, speed;	// Height, background type and speed for the scenery
	unsigned long int init_point;	// Initial point for scenery
	char filename[STRING_LENGTH];	// Filename of the image background, if choosen
} scn_header;

/* Object files header */
typedef struct {
	char version[10];	// Version of the object file
	char image[STRING_LENGTH];	// Image file for object
	char shootimage_file[STRING_LENGTH];	// Image for the object when it shoots
	char upimage_file[STRING_LENGTH];	// Image for the object when it goes up
	char downimage_file[STRING_LENGTH];	// Image for the object when it goes down
	int inmortal;	// True if inmortal
	long int life;	// Object life
	unsigned long int shoot_energy;	// Energy needed for creating this object
	unsigned long int damage;	// Object damage power
	int layer;	// Layer height (draw priority)
	int move_type;	// Object movement behaviour
	char move_file[STRING_LENGTH];	// Movement file specification
	int move_aux;	// Loop enabled if true. It is also the object speed
	int shoot_type;	// Object shoot behaviour
	unsigned long int shoot_count;	// When shoot_type == 4
	char hit_file[STRING_LENGTH];	// Hit object file
	char die_file[STRING_LENGTH];	// Die object file
	char goodies[STRING_LENGTH];	// Goodies object
} obj_header;

/* Event record of a file */
typedef struct {
	int type;	// Event type
	char string[STRING_LENGTH];	// Generic string
	unsigned long int point;	// It will occurs at a point
	unsigned long int aux;	/* Auxiliary variable that can be:
					- Point to jump
					- New Event
					- Vertical position*/
	unsigned long int count; // A counter or the new background type in a background change or object speed
	int x, y;	// Two more variables for positioning
} event_record;

/* Event node for event lists */
typedef struct event_node {
	int type;	// Event type
	char string[STRING_LENGTH];	// Message string
	void *pointer;	// Generic pointer for images and movements
	unsigned long int object_id; // An object id.
	unsigned long int point;	// It will occurs at a point
	unsigned long int aux;	/* Auxiliary variable that can be:
					- Point to jump
					- New Event
					- Vertical position*/
	unsigned long int count; // A counter or the new background type in a background change or 'controled_by' variable
	int x, y;	// Two more variables for positioning
	struct event_node *next;
} event_node;
typedef event_node *event_list;

/* eventpoint_list is a list of all events of a file (directly copied) */
typedef struct eventpoint_node {
	event_record node;
	struct eventpoint_node *next;
} eventpoint_node;
typedef struct eventpoint_node *eventpoint_list;

/* Image node for image lists */
typedef struct image_node {
	png_bytep image;
	png_uint_32 width, height;
	struct image_node *next;
} image_node;
typedef image_node *image_list;

/* Image filename node for image filename list */
typedef struct imagefile_node {
	char filename[STRING_LENGTH];
	image_list image;
	struct imagefile_node *next;
} imagefile_node;
typedef imagefile_node *imagefile_list;

/* imagepoint_list is a list like imagefile but with a 'point' field */
typedef struct imagepoint_node {
	unsigned long int point;
	char filename[STRING_LENGTH];
	png_bytep image;
	png_uint_32 width, height;
	struct imagepoint_node *next;
} imagepoint_node;
typedef struct imagepoint_node *imagepoint_list;

/*
 * Counter list. This is a list created for each jump with a counter 
 * (in events) or when a new human controlled object is created (for
 * initial inmortality) or the left lives of a human
 */
typedef struct counter_node {
	unsigned long int point, count;
	struct counter_node *next;
} counter_node;
typedef counter_node *counter_list;

/* Scenery struct, linked with events list. Like scenery file header ;) */
typedef struct {
	int height, bg_type, speed;
	int vpos;	//	Vertical offset
	unsigned long int current_point;
	image_list back_image;
	event_list next_event;
	counter_list counters;	// Counters for jump events
	event_list events;
} scenery_struct;

/*
 * Object struct, linked with events list. These objects are the template 
 * objects.  
 */
typedef struct {
	image_list image;	// Image pointer for object
	image_list shootimage;	// Image for the object when it shoots
	image_list upimage;	// Image for the object when it goes up
	image_list downimage;	// Image for the object when it goes down
	int inmortal;	// True if inmortal
	long int life;	// Object life
	unsigned long int shoot_energy;	// Energy needed for creating this object
	unsigned long int damage;	// Object damage power
	int layer;	// Layer height for the object (draw priority)
	int move_type;	// Object movement behaviour
	char move_file[STRING_LENGTH];	// This will be a pointer 
	int move_aux;	// Loop enabled. It is also the object speed
	int shoot_type;	// Object shoot behaviour
	unsigned long int shoot_count;	// When shoot_type == 5
	unsigned long int hit_id;	// When something hits the object, other object will be created
	unsigned long int die_id;	// When it die, other object will be created
	unsigned long int goodies;	// New features for other object when it hits this object
	event_list events;
} object_struct;

/* id_list is a list of numbers that identify objects */
typedef struct id_node {
	unsigned long int id;	// Weapon object id.
	int x, y;	// Position offset
	struct id_node *next;
} id_node;
typedef id_node *id_list;

/* Object list. This is the real object, linked to a template object */
typedef struct object_node {
	unsigned long int type;	// Object type, the template
	int x, y;	// Position of the object
	char visible;	// True or false
	unsigned long int controled_by;	// 0 for machine
	long int life;	// Object life
	unsigned long int point; // Current point
	unsigned long int energy;	// Energy acumulated for shoot
	image_list image;	// Object image
	id_list weapons;	// A list of weapons
	struct object_node *father;	// Father of the object
	struct object_node *next;	// Next node
	event_list next_event;
	counter_list counters;	// Counters for jump events
} object_node;
typedef	object_node *object_list;

/* dead_list is a list of objects that must be deleted */
typedef struct dead_node {
	object_list object;
	struct dead_node *next;
} dead_node;
typedef dead_node *dead_list;

typedef struct {
	float x, y, speed;
} float_pos_struct;

#endif /* structs.h */
